!=========================================================================
! Copyright (C) GemTalk Systems 1986-2013.  All Rights Reserved.
!
! $Id: fractio.gs,v 1.11 2008-03-04 22:48:29 bretlb Exp $
!
! Superclass Hierarchy:
!   Fraction, Number, Magnitude, Object.
!
!=========================================================================

removeallmethods Fraction
removeallclassmethods Fraction

category: 'For Documentation Installation only'
classmethod: Fraction
installDocumentation

| doc txt |
doc := GsClassDocumentation newForClass: self.

txt := (GsDocText new) details:
'A Fraction is a Number represented as the ratio of two Integers.'.
doc documentClassWith: txt.

txt := (GsDocText new) details:
'The numerator of the Fraction (an Integer).'.
doc documentInstVar: #numerator with: txt.

txt := (GsDocText new) details:
'The denominator of the Fraction (an Integer).'.
doc documentInstVar: #denominator with: txt.

txt := (GsDocText new) details:
'Do not send class Fraction the message new.  Fractions created in that way are
 meaningless and cannot be handled properly by GemStone''s associative access
 mechanism.  To create a new Fraction, use one of the instance creation methods
 listed here.'.
doc documentClassCategory: #'Instance Creation' with: txt.

self description: doc.
%

category: 'Instance Creation'
classmethod: Fraction
fromStream: aStream

"Returns a Fraction from the stream.  The stream must contain two Integers
 separated by a slash.  (There may be blanks around the slash.)  Generates an
 error if the stream contains anything else, or if an attempt is made to read
 beyond the end of the stream."

| ch n d |

self _checkReadStream: aStream forClass: CharacterCollection.

ch := aStream next.
[ ch isEquivalent: $ ]
whileTrue:
  [ ch := aStream next ].
aStream position: (aStream position - 1).
n := Integer fromStream: aStream.
ch := aStream next.
[ ch isEquivalent: $ ]
whileTrue:
  [ ch := aStream next ].
aStream position: (aStream position - 1).
(aStream next isEquivalent: $/)
ifFalse:[ ^ self _errIncorrectFormat: aStream ].
ch := aStream next.
[ ch isEquivalent: $ ]
whileTrue:
  [ ch := aStream next ].
aStream position: (aStream position - 1).
d := Integer fromStream: aStream.
^ self numerator: n denominator: d
%

category: 'Instance Creation'
classmethod: Fraction
numerator: numInt denominator: denomInt

"Returns an instance of Fraction with numerator numInt and denominator
 denomInt.  If that Fraction can be reduced, this method returns the
 corresponding Integer instead.  The result is made invariant.

 If either argument (numerator or denominator) is not an Integer, that
 argument is truncated to the corresponding Integer."

numInt _validateClass: Number.
denomInt _validateClass: Number.

^ (self new _setNumerator: numInt denominator: denomInt) 
    _reduce immediateInvariant
%

category: 'Accessing'
method: Fraction
at: anIndex

"Disallowed."

self shouldNotImplement: #at:
%

category: 'Accessing'
method: Fraction
at: anIndex put: aNumber

"Disallowed.  You may not change the value of a Fraction."

self shouldNotImplement: #at:put:
%

category: 'Accessing'
method: Fraction
denominator

"Returns the denominator of the receiver."

^denominator
%

category: 'Accessing'
method: Fraction
instVarAt: anIndex put: aValue

"Disallowed.  You may not change the value of a Fraction."

self shouldNotImplement: #instVarAt:put:
%

category: 'Accessing'
method: Fraction
numerator

"Returns the numerator of the receiver."

^numerator
%

category: 'Accessing'
method: Fraction
size: anInteger

"Disallowed.  You may not change the size of a Fraction."

self shouldNotImplement: #size:
%

category: 'Private'
method: Fraction
_setNumerator: n denominator: d

"If either the numerator or denominator is not an Integer, this method performs
 the truncation.  It then assigns the numerator and denominator.  

 The receiver must still be variant.  The sender is responsible for making the
 receiver invariant."

| denom |
denom := d truncated abs.
(denom = 0) ifTrue: [
  ^ n _errorDivideByZero 
  ].
numerator := n truncated.
denominator :=  denom .
(d < 0) ifTrue: [numerator := numerator negated] .

%

category: 'Private'
method: Fraction
_reduce

"reduce to form using least common denominator."

| gcd |

(numerator == 0) ifTrue: [ ^0 ].
gcd := numerator gcd: denominator.
numerator := numerator // gcd.
denominator := denominator // gcd.
(denominator == 1) ifTrue:[ 
  ^ numerator 
  ].
^ self 
%

category: 'Formatting'
method: Fraction
asString

"Returns a String of the form 'numerator/denominator'."

| result |
result := numerator asString .
result add: $/ ; addAll: denominator asString .
^ result
%

category: 'Converting'
method: Fraction
asScaledDecimal: scale

"Returns a ScaledDecimal representation of the receiver."

numerator == nil ifTrue:[ ^ PlusSignalingNaN ] .
denominator == nil ifTrue:[ ^ PlusSignalingNaN ] .
^ ScaledDecimal numerator: numerator denominator: denominator scale: scale.
%

! asFloat moved to fractio2.gs 

category: 'Converting'
method: Fraction
asDecimalFloat

"Returns an instance of DecimalFloat that has the value of the receiver."

numerator == nil ifTrue:[ ^ DecimalPlusSignalingNaN ] .
denominator == nil ifTrue:[ ^ DecimalPlusSignalingNaN ] .
^ numerator asDecimalFloat / denominator asDecimalFloat
%

category: 'Converting'
method: Fraction
asFraction

"(R) Returns the receiver."

^self
%

category: 'Converting'
method: Fraction
_coerce: aNumber

"Reimplemented from Number."

^aNumber asFraction
%

category: 'Converting'
method: Fraction
_generality

"Reimplemented from Number."

^60
%

category: 'Comparing'
method: Fraction
= aFraction

"Returns true if the receiver is equal to aFraction; returns false otherwise."

| fracNumerator fracDenominator |

((numerator == nil) _or: [denominator == nil]) ifTrue: [^ false ].

(aFraction _class == Fraction) 
  ifFalse: [ ^ self _retry: #= coercing: aFraction ].

fracNumerator := aFraction numerator.
fracDenominator := aFraction denominator.
((fracNumerator == nil) _or: [fracDenominator == nil]) ifTrue: [^ false ].

^ ((numerator = fracNumerator) _and: [denominator = fracDenominator])
%

category: 'Comparing'
method: Fraction
~= aFraction

"Returns true if the receiver is not equal to aFraction; returns false
 otherwise."

| fracNumerator fracDenominator |

((numerator == nil) _or: [denominator == nil]) ifTrue: [^ false ].

(aFraction _class == Fraction) 
  ifFalse: [ ^ self _retry: #~= coercing: aFraction ].

fracNumerator := aFraction numerator.
fracDenominator := aFraction denominator.
((fracNumerator == nil) _or: [fracDenominator == nil]) ifTrue: [^ false ].

^ ((numerator ~= fracNumerator) _or: [denominator ~= fracDenominator])
%

category: 'Comparing'
method: Fraction
>= aFraction

"Returns true if the receiver is greater than aFraction; returns false
 otherwise."

| fracNumerator fracDenominator |

((numerator == nil) _or: [denominator == nil]) ifTrue: [^ false ].

(aFraction _class == Fraction) 
  ifFalse: [ ^ self _retry: #>= coercing: aFraction ].

fracNumerator := aFraction numerator.
fracDenominator := aFraction denominator.
((fracNumerator == nil) _or: [fracDenominator == nil]) ifTrue: [^ false ].

^ ((numerator * fracDenominator) >= (denominator * fracNumerator))
%

category: 'Comparing'
method: Fraction
> aFraction

"Returns true if the receiver is greater than aFraction; returns false
 otherwise."

| fracNumerator fracDenominator |

((numerator == nil) _or: [denominator == nil]) ifTrue: [^ false ].

(aFraction _class == Fraction) 
  ifFalse: [ ^ self _retry: #> coercing: aFraction ].

fracNumerator := aFraction numerator.
fracDenominator := aFraction denominator.
((fracNumerator == nil) _or: [fracDenominator == nil]) ifTrue: [^ false ].

^ ((numerator * fracDenominator) > (denominator * fracNumerator))
%

category: 'Comparing'
method: Fraction
<= aFraction

"Returns true if the receiver is less than or equal to aFraction; returns false
 otherwise."

| fracNumerator fracDenominator |

((numerator == nil) _or: [denominator == nil]) ifTrue: [^ false ].

(aFraction _class == Fraction) 
  ifFalse: [ ^ self _retry: #<= coercing: aFraction ].

fracNumerator := aFraction numerator.
fracDenominator := aFraction denominator.
((fracNumerator == nil) _or: [fracDenominator == nil]) ifTrue: [^ false ].

^ ((numerator * fracDenominator) <= (denominator * fracNumerator))
%

category: 'Comparing'
method: Fraction
< aFraction

"Returns true if the receiver is less than aFraction; returns false otherwise."

| fracNumerator fracDenominator |

((numerator == nil) _or: [denominator == nil]) ifTrue: [^ false ].

(aFraction _class == Fraction) 
  ifFalse: [ ^ self _retry: #< coercing: aFraction ].

fracNumerator := aFraction numerator.
fracDenominator := aFraction denominator.
((fracNumerator == nil) _or: [fracDenominator == nil]) ifTrue: [^ false ].

^ ((numerator * fracDenominator) < (denominator * fracNumerator))
%

! hash inherited from Number

category: 'Truncation and Rounding'
method: Fraction
truncated

"Returns the integer that is closest to the receiver, on the same side
 of the receiver as zero is located."

^numerator quo: denominator
%

category: 'Arithmetic'
method: Fraction
* aFraction

"Returns the result of multiplying the receiver by aFraction."

| fracNumerator fracDenominator |

(aFraction _class == Fraction) 
  ifFalse: [ ^ self _retry: #* coercing: aFraction ].

((numerator == nil) _or: [denominator == nil]) ifTrue: [^ Fraction new ].

fracNumerator := aFraction numerator.
fracDenominator := aFraction denominator.
((fracNumerator == nil) _or: [fracDenominator == nil]) 
  ifTrue: [^ Fraction new ].

^(Fraction
    numerator: numerator * fracNumerator
    denominator: denominator * fracDenominator)
%

category: 'Arithmetic'
method: Fraction
+ aFraction

"Returns the sum of the receiver and aFraction."

| fracNumerator fracDenominator commonDenominator newNumerator |

(aFraction _class == Fraction) 
  ifFalse: [ ^ self _retry: #+ coercing: aFraction ].

((numerator == nil) _or: [denominator == nil]) ifTrue: [^ Fraction new ].

fracNumerator := aFraction numerator.
fracDenominator := aFraction denominator.
((fracNumerator == nil) _or: [fracDenominator == nil]) 
  ifTrue: [^ Fraction new ].

(denominator = fracDenominator)
  ifTrue: [ ^(Fraction
                numerator: numerator + aFraction numerator
                denominator: denominator)].

commonDenominator := denominator lcm: fracDenominator.
newNumerator := numerator
                   * (commonDenominator / denominator)
                   + (aFraction numerator *
                       (commonDenominator / fracDenominator)).
 ^(Fraction
     numerator: newNumerator
     denominator: commonDenominator)
%

category: 'Arithmetic'
method: Fraction
- aFraction

"Returns the difference between the receiver and aFraction."

| fracNumerator fracDenominator commonDenominator newNumerator |

(aFraction _class == Fraction) 
  ifFalse: [ ^ self _retry: #- coercing: aFraction ].

((numerator == nil) _or: [denominator == nil]) ifTrue: [^ Fraction new ].

fracNumerator := aFraction numerator.
fracDenominator := aFraction denominator.
((fracNumerator == nil) _or: [fracDenominator == nil]) 
  ifTrue: [^ Fraction new ].

(denominator = fracDenominator)
  ifTrue: [ ^(Fraction
                numerator: numerator - aFraction numerator
                denominator: denominator)].

commonDenominator := denominator lcm: fracDenominator.
newNumerator := numerator
                   * (commonDenominator / denominator)
                   - (aFraction numerator *
                       (commonDenominator / fracDenominator)).
 ^(Fraction
     numerator: newNumerator
     denominator: commonDenominator)
%

category: 'Arithmetic'
method: Fraction
/ aFraction

"Returns the result of dividing the receiver by aFraction."

| fracNumerator fracDenominator |

(aFraction _class == Fraction) 
  ifFalse: [ ^ self _retry: #/ coercing: aFraction ].

((numerator == nil) _or: [denominator == nil]) ifTrue: [^ Fraction new ].

fracNumerator := aFraction numerator.
fracDenominator := aFraction denominator.
((fracNumerator == nil) _or: [fracDenominator == nil]) 
  ifTrue: [^ Fraction new ].

(fracDenominator == 0) ifTrue: [ ^ aFraction _errorDivideByZero ].
^(Fraction
    numerator: numerator * fracDenominator
    denominator: denominator * aFraction numerator)
%

category: 'Arithmetic'
method: Fraction
negated

"Returns a Number that is the negation of the receiver."

^self class numerator: numerator negated denominator: denominator
%

category: 'Arithmetic'
method: Fraction
reciprocal

"Returns 1 divided by the receiver.  Generates an error if the receiver is 0."

(numerator == 0) ifTrue: [ ^ self _errorDivideByZero].
(numerator == 1) ifTrue: [^denominator].
(numerator == -1) ifTrue: [^denominator negated].
^self class numerator: denominator denominator: numerator
%

category: 'Storing and Loading'
classmethod: Fraction
loadFrom: passiveObj

"Reads from passiveObj the passive form of an object.  Converts the object to
 its active form by loading the information into a new instance of the receiver.
 Returns the new instance."

| inst num den marker |

marker := passiveObj objectPositionMarker. 

passiveObj readNamedIV.
num := passiveObj ivValue.
passiveObj readNamedIV.
den := passiveObj ivValue.
  
passiveObj skipNamedInstVars.

inst := self numerator: num denominator: den.
passiveObj hasRead: inst marker: marker.
^inst.
%

category: 'Storing and Loading'
method: Fraction
writeTo: passiveObj

"Converts the receiver to its passive form and writes that information on
 passiveObj."

"Reimplemented from Number since the receiver has a non-literal representation."

^super basicWriteTo: passiveObj
%

category: 'Testing'
method: Fraction
even

"(R) Returns true if the receiver is an even integer, false otherwise."
^ false.
%

category: 'Testing'
method: Fraction
odd

"(R) Returns true if the receiver is an odd integer, false otherwise."
^ false.
%
